#!/usr/bin/env python3
"""
Gui module to build and save a simulation set
"""

import os
import json
from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QGridLayout,
    QLabel,
    QPushButton,
    QFileDialog,
    QSpinBox,
    QDoubleSpinBox,
    QCheckBox,
    QListWidget,
    QTableWidget,
    QHeaderView,
    QTableWidgetItem,
    QGroupBox,
    QRadioButton,
)
from PyQt5 import QtCore
from Bio import SeqIO
import pandas as pd


class Gui:
    """
    class to generate GUI.
    """

    def __init__(self):
        self.genes_dict = {}
        self.stop_condition = ()

        self.app = QApplication([])
        self.window = QWidget()
        self.window.resize(100, 800)  # initial window size
        grid = QGridLayout()
        self.window.setLayout(grid)

        self.window.setWindowTitle("Simulation Builder")

        grid.addWidget(QLabel("concentration file: "), 0, 0)
        selected_concentration_label = QLabel("      ")
        selected_concentration_label.setObjectName("selected_concentration_label")
        grid.addWidget(selected_concentration_label, 0, 1)

        concentration_file_open_button = QPushButton("Open file")
        concentration_file_open_button.setToolTip("Open concentration file")
        concentration_file_open_button.clicked.connect(self.open_concentrations_file)
        grid.addWidget(concentration_file_open_button, 0, 2)

        grid.addWidget(QLabel("Pre-populate: "), 1, 0)
        pre_populate_check_box = QCheckBox()
        pre_populate_check_box.setObjectName("pre_populate_check_box")
        grid.addWidget(pre_populate_check_box, 1, 1)

        grid.addWidget(QLabel("Add FASTA file :"), 2, 0)
        add_fasta_file_button = QPushButton("Add FASTA file")
        add_fasta_file_button.setToolTip(
            "Reads FASTA file and make its genes available for simulation"
        )
        add_fasta_file_button.clicked.connect(self.open_fasta_file)
        grid.addWidget(add_fasta_file_button, 2, 1)

        load_batch_file_button = QPushButton("Load batch CSV file")
        load_batch_file_button.setToolTip(
            "load a CSV file with a list of Genes, copy number, initiation and termination rates."
        )
        load_batch_file_button.clicked.connect(self.load_batch_file)
        grid.addWidget(load_batch_file_button, 2, 3)

        grid.addWidget(QLabel("Loaded FASTA files:"), 3, 0)
        fasta_files_listbox = QListWidget()
        fasta_files_listbox.setObjectName("fasta_files_listbox")
        fasta_files_listbox.clicked.connect(self.onselect_fasta_file)
        grid.addWidget(fasta_files_listbox, 3, 1)

        grid.addWidget(QLabel("Genes:"), 3, 2)
        genes_listbox = QListWidget()
        genes_listbox.setObjectName("genes_listbox")
        grid.addWidget(genes_listbox, 3, 3)

        rates_groupbox = QGroupBox()
        rates_groupbox_grid = QGridLayout()
        rates_groupbox.setLayout(rates_groupbox_grid)
        grid.addWidget(rates_groupbox, 3, 4)
        rates_groupbox_grid.addWidget(QLabel("Initiation rate: "), 0, 0)
        init_rate_spinbox = QDoubleSpinBox()
        init_rate_spinbox.setObjectName("init_rate_spinbox")
        init_rate_spinbox.setRange(0.01, 100.00)
        init_rate_spinbox.setSingleStep(0.01)
        rates_groupbox_grid.addWidget(init_rate_spinbox, 0, 1)

        rates_groupbox_grid.addWidget(QLabel("Termination rate: "), 1, 0)
        term_rate_spinbox = QDoubleSpinBox()
        term_rate_spinbox.setObjectName("term_rate_spinbox")
        term_rate_spinbox.setRange(0.01, 100.00)
        term_rate_spinbox.setSingleStep(0.01)
        rates_groupbox_grid.addWidget(term_rate_spinbox, 1, 1)

        rates_groupbox_grid.addWidget(QLabel("Transcript copy number: "), 2, 0)
        gene_copy_number_spinbox = QDoubleSpinBox()
        genes_listbox.setSelectionMode(2)
        gene_copy_number_spinbox.setObjectName("gene_copy_number_spinbox")
        gene_copy_number_spinbox.setRange(1, 1000)
        gene_copy_number_spinbox.setSingleStep(1)
        rates_groupbox_grid.addWidget(gene_copy_number_spinbox, 2, 1)

        add_gene_button = QPushButton("Add gene")
        add_gene_button.clicked.connect(self.add_simulation_entry)
        grid.addWidget(add_gene_button, 4, 1, 1, 4)

        grid.addWidget(QLabel("Added simulations: "), 5, 0)
        added_simulations_listbox = QTableWidget()
        added_simulations_listbox.setObjectName("added_simulations_listbox")
        added_simulations_listbox.setColumnCount(5)
        added_simulations_listbox.setShowGrid(False)
        added_simulations_listbox.setHorizontalHeaderLabels(
            [
                "Fasta file",
                "Gene",
                "Initiation\nRate",
                "Termination\nRate",
                "Transcript\nCopy Number",
            ]
        )
        added_simulations_listbox.resizeColumnsToContents()
        added_simulations_listbox.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch
        )
        added_simulations_listbox.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch
        )
        grid.addWidget(added_simulations_listbox, 6, 0, 6, 6)

        remove_entry_button = QPushButton("Remove entry")
        remove_entry_button.clicked.connect(self.remove_simulation_entry)
        grid.addWidget(remove_entry_button, 14, 5)

        termination_condition_groupbox = QGroupBox()
        termination_condition_groupbox.setTitle("Stop condition")
        termination_condition_groupbox_grid = QGridLayout()
        termination_condition_groupbox.setLayout(termination_condition_groupbox_grid)
        grid.addWidget(termination_condition_groupbox, 15, 0, 2, 2)

        iteration_limit_radiobutton = QRadioButton("Iteration limit:")

        steady_state_checkbox = QCheckBox("After steady state: ")
        steady_state_checkbox.setObjectName("steady_state_checkbox")
        termination_condition_groupbox_grid.addWidget(steady_state_checkbox, 0, 0)
        steady_state_checkbox.clicked.connect(self.changed_steady_state_checkbox)
        # termination_condition_groupbox_grid.addWidget(steady_state_checkbox, 0, 1)

        iteration_limit_radiobutton.setObjectName("iteration_limit_radiobutton")
        termination_condition_groupbox_grid.addWidget(iteration_limit_radiobutton, 1, 0)
        iteration_limit_spinbox = QSpinBox()
        iteration_limit_spinbox.setObjectName("iteration_limit_spinbox")
        iteration_limit_spinbox.setRange(0, 2147483647)
        iteration_limit_spinbox.valueChanged.connect(self.changed_iteration_limit)
        termination_condition_groupbox_grid.addWidget(iteration_limit_spinbox, 1, 1)

        time_limit_radiobutton = QRadioButton("Time limit:")
        time_limit_radiobutton.setObjectName("time_limit_radiobutton")
        time_limit_radiobutton.setChecked(True)  # have a checked option for starters.
        termination_condition_groupbox_grid.addWidget(time_limit_radiobutton, 2, 0)
        time_limit_spinbox = QDoubleSpinBox()
        time_limit_spinbox.setObjectName("time_limit_spinbox")
        time_limit_spinbox.setRange(0, 10e10)
        time_limit_spinbox.setValue(60)  # set a default value
        self.stop_condition = ("time", float(time_limit_spinbox.value()))
        time_limit_spinbox.valueChanged.connect(self.changed_time_limit_entry)
        termination_condition_groupbox_grid.addWidget(time_limit_spinbox, 2, 1)

        finished_ribosomes_limit_radiobutton = QRadioButton("Terminated Ribosomes:")
        finished_ribosomes_limit_radiobutton.setObjectName(
            "finished_ribosomes_limit_radiobutton"
        )
        termination_condition_groupbox_grid.addWidget(
            finished_ribosomes_limit_radiobutton, 3, 0
        )
        finished_ribosomes_spinbox = QSpinBox()
        finished_ribosomes_spinbox.setObjectName("finished_ribosomes_spinbox")
        finished_ribosomes_spinbox.setRange(0, 2147483647)
        finished_ribosomes_spinbox.valueChanged.connect(self.changed_finished_ribosomes)
        termination_condition_groupbox_grid.addWidget(finished_ribosomes_spinbox, 3, 1)

        history_groupbox = QGroupBox()
        history_groupbox.setFlat(True)
        history_groupbox_grid = QGridLayout()
        history_groupbox.setLayout(history_groupbox_grid)
        grid.addWidget(history_groupbox, 15, 3, 2, 1)

        history_groupbox_grid.addWidget(QLabel("Number of history entries: "), 0, 0)
        history_size_spinbox = QSpinBox()
        history_size_spinbox.setObjectName("history_size_spinbox")
        history_size_spinbox.setRange(1, 2147483647)
        history_size_spinbox.setValue(100000)
        history_groupbox_grid.addWidget(history_size_spinbox, 0, 1)

        generate_simulation_button = QPushButton("Generate Simulation file")
        generate_simulation_button.clicked.connect(self.generate_simulation_file)
        grid.addWidget(generate_simulation_button, 17, 1, 3, 4)
        self.show()

    def show(self):
        """
        bring GUI window to front and show it.
        """
        self.window.show()
        self.app.exec_()

    def open_concentrations_file(self):
        """
        open dialog to select concentration file.
        """
        concentrations_file, _ = QFileDialog.getOpenFileName(
            self.window, "Select file", os.getcwd(), "CSV File (*.csv )"
        )
        if concentrations_file == "":
            return
        selected_concentration_label = self.window.findChild(
            QLabel, "selected_concentration_label"
        )
        selected_concentration_label.setText(concentrations_file)

    def changed_steady_state_checkbox(self, state):
        steady_state_checkbox = self.window.findChild(
            QCheckBox, "steady_state_checkbox"
        )
        iteration_limit_radiobutton = self.window.findChild(
            QRadioButton, "iteration_limit_radiobutton"
        )
        iteration_limit_spinbox = self.window.findChild(
            QSpinBox, "iteration_limit_spinbox"
        )
        time_limit_radiobutton = self.window.findChild(
            QRadioButton, "time_limit_radiobutton"
        )
        if iteration_limit_radiobutton.isChecked():
            time_limit_radiobutton.setChecked(True)
        iteration_limit_radiobutton.setChecked(False)
        iteration_limit_radiobutton.setEnabled(not steady_state_checkbox.checkState())
        iteration_limit_spinbox.setEnabled(not steady_state_checkbox.checkState())
        # update termination condition
        finished_ribosomes_limit_radiobutton = self.window.findChild(
            QRadioButton, "finished_ribosomes_limit_radiobutton"
        )
        if iteration_limit_radiobutton.isChecked():
            self.changed_iteration_limit()
        elif time_limit_radiobutton.isChecked():
            self.changed_time_limit_entry()
        else:
            self.changed_finished_ribosomes()
        return

    def open_fasta_file(self):
        """
        Open dialog to select FASTA file with genes to simulate.
        Also populate gene box.
        """
        fasta_file_chooser, _ = QFileDialog.getOpenFileName(
            self.window,
            "Select file",
            os.getcwd(),
            "All compatible files (*.txt *.fasta *.fna *.fa);;txt File (*.txt );;FASTA file (*.fasta);;FNA file (*.fna *.fa)",
        )
        if fasta_file_chooser == "" or fasta_file_chooser in self.genes_dict.keys():
            return
        self.genes_dict[fasta_file_chooser] = [
            record.id for record in SeqIO.parse(fasta_file_chooser, "fasta")
        ]
        # update fasta files listbox
        fasta_files_listbox = self.window.findChild(QListWidget, "fasta_files_listbox")
        if fasta_files_listbox.count() > 0:
            fasta_files_listbox.clear()
        fasta_files_listbox.addItems(self.genes_dict.keys())
        fasta_files_listbox.setCurrentRow(fasta_files_listbox.count() - 1)
        # update genes listbox
        genes_listbox = self.window.findChild(QListWidget, "genes_listbox")
        if genes_listbox.count() > 0:
            genes_listbox.clear()
        genes_listbox.addItems(self.genes_dict[fasta_file_chooser])
        genes_listbox.setCurrentRow(genes_listbox.count())

    def load_batch_file(self):
        """
        Open dialog to select CSV file with ORF, transcript copy number, initiation and termination rates.
        """
        batch_file_path, _ = QFileDialog.getOpenFileName(
            self.window, "Select file", os.getcwd(), "CSV File (*.csv )"
        )
        added_simulations_listbox = self.window.findChild(
            QTableWidget, "added_simulations_listbox"
        )
        if batch_file_path == "":
            return
        df = pd.read_csv(batch_file_path)
        if not set(["ORF", "RNA_copies", "Ini_rate", "Term_rate"]).issubset(df.columns):
            return
        for _, row in df.iterrows():
            for key in self.genes_dict.keys():
                index_to_update = added_simulations_listbox.rowCount()
                if row.loc["ORF"] in self.genes_dict[key]:
                    items = added_simulations_listbox.findItems(
                        row.loc["ORF"], QtCore.Qt.MatchExactly
                    )
                    if items:
                        index_to_update = items[0].row()  # go to the row to edit entry.
                    else:
                        added_simulations_listbox.insertRow(
                            index_to_update
                        )  # create new entry
                added_simulations_listbox.setItem(
                    index_to_update, 0, QTableWidgetItem(key)
                )
                added_simulations_listbox.setItem(
                    index_to_update, 1, QTableWidgetItem(row.loc["ORF"])
                )
                added_simulations_listbox.setItem(
                    index_to_update, 2, QTableWidgetItem(str(row.loc["Ini_rate"]))
                )
                added_simulations_listbox.setItem(
                    index_to_update, 3, QTableWidgetItem(str(row.loc["Term_rate"]))
                )
                added_simulations_listbox.setItem(
                    index_to_update, 4, QTableWidgetItem(str(row.loc["RNA_copies"]))
                )

    def onselect_fasta_file(self):
        """
        Update gene box with the genes of the selected fasta file.
        """
        fasta_files_listbox = self.window.findChild(QListWidget, "fasta_files_listbox")
        value = fasta_files_listbox.currentItem().text()
        # update genes listbox
        genes_listbox = self.window.findChild(QListWidget, "genes_listbox")
        if genes_listbox.count() > 0:
            genes_listbox.clear()
        genes_listbox.addItems(self.genes_dict[value])
        genes_listbox.setCurrentRow(genes_listbox.count())

    def get_entries_to_add_and_update(self, selected_fasta_file, selected_genes):
        added_simulations_listbox = self.window.findChild(
            QTableWidget, "added_simulations_listbox"
        )
        fasta_files_list = [
            str(added_simulations_listbox.item(i, 0).text())
            for i in range(added_simulations_listbox.rowCount())
        ]
        genes_list = [
            str(added_simulations_listbox.item(i, 1).text())
            for i in range(added_simulations_listbox.rowCount())
        ]

        fasta_files_list_indexes = [
            i for i, x in enumerate(fasta_files_list) if x == selected_fasta_file
        ]
        genes_list_indexes_to_be_updated = [
            i for i, x in enumerate(genes_list) if x in selected_genes
        ]
        genes_list_to_be_added = [
            i for i, x in enumerate(selected_genes) if x not in genes_list
        ]
        indexes_to_update = set(fasta_files_list_indexes).intersection(
            genes_list_indexes_to_be_updated
        )
        return genes_list_to_be_added, indexes_to_update

    def add_simulation_entry(self):
        """
        Use the selected fasta file, gene and gene copy number to add a new simulation entry.
        If fasta file AND gene is already added, update that entry with new copy number.
        """
        # we need: a gene and copy number.
        genes_listbox = self.window.findChild(QListWidget, "genes_listbox")
        fasta_files_listbox = self.window.findChild(QListWidget, "fasta_files_listbox")
        added_simulations_listbox = self.window.findChild(
            QTableWidget, "added_simulations_listbox"
        )
        gene_copy_number_spinbox = self.window.findChild(
            QDoubleSpinBox, "gene_copy_number_spinbox"
        )
        initiation_rate_spinbox = self.window.findChild(
            QDoubleSpinBox, "init_rate_spinbox"
        )
        termination_rate_spinbox = self.window.findChild(
            QDoubleSpinBox, "term_rate_spinbox"
        )
        if (
            len(genes_listbox.selectedItems()) == 0
            or len(fasta_files_listbox.selectedItems()) != 1
        ):
            return
        selected_fasta_file = fasta_files_listbox.currentItem().text()
        selected_genes = [item.text() for item in genes_listbox.selectedItems()]
        genes_list_to_be_added, indexes_to_update = \
            self.get_entries_to_add_and_update(
                selected_fasta_file, selected_genes
            )

        if len(indexes_to_update) > 0:
            # update gene copy number, initiation rate and termination rate
            for i in indexes_to_update:
                added_simulations_listbox.setItem(
                    i, 2, QTableWidgetItem(str(initiation_rate_spinbox.value()))
                )
                added_simulations_listbox.setItem(
                    i, 3, QTableWidgetItem(str(termination_rate_spinbox.value()))
                )
                added_simulations_listbox.setItem(
                    i, 4, QTableWidgetItem(str(gene_copy_number_spinbox.value()))
                )
        # process new entries
        for selected_gene_index in genes_list_to_be_added:
            # new entry
            indexes_to_update = added_simulations_listbox.rowCount()
            added_simulations_listbox.insertRow(indexes_to_update)
            added_simulations_listbox.setItem(
                indexes_to_update, 0, QTableWidgetItem(selected_fasta_file)
            )
            added_simulations_listbox.setItem(
                indexes_to_update,
                1,
                QTableWidgetItem(selected_genes[selected_gene_index]),
            )
            added_simulations_listbox.setItem(
                indexes_to_update,
                2,
                QTableWidgetItem(str(initiation_rate_spinbox.value())),
            )
            added_simulations_listbox.setItem(
                indexes_to_update,
                3,
                QTableWidgetItem(str(termination_rate_spinbox.value())),
            )

            added_simulations_listbox.setItem(
                indexes_to_update,
                4,
                QTableWidgetItem(str(gene_copy_number_spinbox.value())),
            )
        return

    def remove_simulation_entry(self):
        """
        Remove selected simulation entry.
        """
        added_simulations_listbox = self.window.findChild(
            QTableWidget, "added_simulations_listbox"
        )
        index = added_simulations_listbox.currentRow()
        added_simulations_listbox.removeRow(index)

    def changed_iteration_limit(self):
        """
        select iteration limit as the user changed its value.
        """
        iteration_limit_radiobutton = self.window.findChild(
            QRadioButton, "iteration_limit_radiobutton"
        )
        iteration_limit_spinbox = self.window.findChild(
            QSpinBox, "iteration_limit_spinbox"
        )
        iteration_limit_radiobutton.setChecked(True)
        self.stop_condition = ("iteration", int(iteration_limit_spinbox.value()))

    def changed_time_limit_entry(self):
        """
        select time limit as the user changed its value.
        """
        time_limit_radiobutton = self.window.findChild(
            QRadioButton, "time_limit_radiobutton"
        )
        time_limit_spinbox = self.window.findChild(QDoubleSpinBox, "time_limit_spinbox")
        steady_state_checkbox = self.window.findChild(
            QCheckBox, "steady_state_checkbox"
        )
        time_limit_radiobutton.setChecked(True)
        if steady_state_checkbox.checkState():
            self.stop_condition = (
                "steady_state_time",
                float(time_limit_spinbox.value()),
            )
        else:
            self.stop_condition = ("time", float(time_limit_spinbox.value()))

    def changed_finished_ribosomes(self):
        """
        select finished ribosomes as the user changed its value.
        """
        finished_ribosomes_limit_radiobutton = self.window.findChild(
            QRadioButton, "finished_ribosomes_limit_radiobutton"
        )
        finished_ribosomes_spinbox = self.window.findChild(
            QSpinBox, "finished_ribosomes_spinbox"
        )
        steady_state_checkbox = self.window.findChild(
            QCheckBox, "steady_state_checkbox"
        )
        finished_ribosomes_limit_radiobutton.setChecked(True)
        if steady_state_checkbox.checkState():
            self.stop_condition = (
                "steady_state_ribosomes",
                int(finished_ribosomes_spinbox.value()),
            )
        else:
            self.stop_condition = ("ribosomes", int(finished_ribosomes_spinbox.value()))

    def generate_simulation_file(self):
        """
        This method should assemble the json and ask the user for a place to save the file.
        """
        file_name = QFileDialog.getSaveFileName(
            self.window,
            "Save simulation configuration",
            os.getcwd(),
            "Simulation Files (*.json) ;;All files(*)",
            options=QFileDialog.DontUseNativeDialog,
        )
        if file_name == "":
            return
        if file_name[0] and file_name[1].split(".")[1] in file_name[0].split(".")[-1]:
            file_name = file_name[0]  # file name already has extension.
        else:
            file_name = file_name[0] + "." + file_name[1].split(".")[1][:-1]
        selected_concentration_label = self.window.findChild(
            QLabel, "selected_concentration_label"
        )
        pre_populate_check_box = self.window.findChild(
            QCheckBox, "pre_populate_check_box"
        )
        history_size_spinbox = self.window.findChild(QSpinBox, "history_size_spinbox")
        added_simulations_table = self.window.findChild(
            QTableWidget, "added_simulations_listbox"
        )

        sb = SimulationBuilder()
        sb.set_concentration_file(selected_concentration_label.text())
        sb.set_pre_populate(pre_populate_check_box.checkState())
        sb.set_history_size(int(history_size_spinbox.value()))
        sb.set_halt_condition(*self.stop_condition)
        for index in range(added_simulations_table.rowCount()):
            fasta_file = added_simulations_table.item(index, 0).text()
            gene = added_simulations_table.item(index, 1).text()
            initiation_rate = float(added_simulations_table.item(index, 2).text())
            termination_rate = float(added_simulations_table.item(index, 3).text())
            transcript_copy_number = float(
                added_simulations_table.item(index, 4).text()
            )
            sb.add_mRNA_entry(
                fasta_file,
                gene,
                initiation_rate,
                termination_rate,
                transcript_copy_number,
            )
        sb.save_simulation(file_name)

        return


class SimulationBuilder:
    """
    class to generate JSON config file
    """

    def __init__(self):
        """
        Create fields with default values.
        """
        self.data = {
            "concentration_file": "",
            "pre_populate": True,
            "mRNA_entries": [],
            "history_size": 100000,
        }

    def set_concentration_file(self, file_path):
        self.data["concentration_file"] = file_path

    def set_pre_populate(self, pre_populate):
        self.data["pre_populate"] = pre_populate

    def set_history_size(self, history_size):
        if history_size > 0:
            self.data["history_size"] = history_size

    def add_mRNA_entry(
        self,
        fasta_file,
        gene,
        initiation_rate,
        termination_rate,
        transcript_copy_number,
    ):
        if initiation_rate < 0 or termination_rate < 0 or transcript_copy_number <= 0:
            return
        entry = {
            "fasta_file": fasta_file,
            "gene": gene,
            "initiation_rate": initiation_rate,
            "termination_rate": termination_rate,
            "transcript_copy_number": transcript_copy_number,
        }
        if entry not in self.data["mRNA_entries"]:
            self.data["mRNA_entries"].append(entry)

    def set_halt_condition(self, condition, value):
        if value < 0:
            return
        if condition.lower() == "iteration":
            self.data["iteration_limit"] = int(value)
        elif condition.lower() == "time":
            self.data["time_limit"] = value
        elif condition.lower() == "ribosomes":
            self.data["finished_ribosomes"] = int(value)
        elif condition.lower() == "steady_state_time":
            self.data["steady_state_time"] = value
        elif condition.lower() == "steady_state_ribosomes":
            self.data["steady_state_ribosomes"] = value

    def save_simulation(self, file_path):
        # validation
        if (
            len(
                set(
                    [
                        "iteration_limit",
                        "time_limit",
                        "finished_ribosomes",
                        "steady_state_time",
                        "steady_state_ribosomes",
                    ]
                ).intersection(self.data.keys())
            )
            == 0
        ):
            return
        if len(self.data["mRNA_entries"]) == 0:
            return
        # convert absolute paths to relative path
        base_path = os.path.split(file_path)[0] + os.sep
        self.data["concentration_file"] = os.path.relpath(
            self.data["concentration_file"], start=base_path
        )
        for mRNA_entry in self.data["mRNA_entries"]:
            mRNA_entry["fasta_file"] = os.path.relpath(
                mRNA_entry["fasta_file"], start=base_path
            )
        # valid simulation. save.
        with open(file_path, "w", encoding="utf-8") as outfile:
            json.dump(self.data, outfile, indent=4)


if __name__ == "__main__":
    Gui()



#!/usr/bin/env python3
import sys
import os
import json
from dataclasses import dataclass
from PyQt5.QtWidgets import (
    QApplication,
    QGridLayout,
    QTreeWidget,
    QTreeWidgetItem,
    QWidget,
    QMainWindow,
    QDialog,
    QLabel,
    QDialogButtonBox,
    QLineEdit,
    QMessageBox,
    QPushButton
)
from PyQt5 import QtCore


@dataclass
class PairingRelationship:
    codon: str
    anticodons: list[str]

    @classmethod
    def load_dict_one_position(cls, json_file_name: str, pairing_type: str) -> list['PairingRelationship']:
        result = []
        with open(json_file_name, encoding="utf-8") as json_file:
            data = json.load(json_file)
            for codon in data[pairing_type]:
                result.append(cls(codon, data[pairing_type][codon]))
            return result

    @classmethod
    def convert_one_position_from_dict(cls, data: dict, pairing_type: str) -> list['PairingRelationship']:
        result = []
        for codon in data[pairing_type]:
            result.append(cls(codon, data[pairing_type][codon]))
        return result

    @classmethod
    def load_all_dict(cls, json_file_name: str):
        result = {}
        if os.path.isfile(json_file_name) and os.path.exists(json_file_name):
            # this is an existing file. Edit:
            with open(json_file_name, encoding="utf-8") as json_file:
                data = json.load(json_file)
                for pairing_type in data.keys():
                    if pairing_type == "Pairing Rules":
                        result[pairing_type] = data[pairing_type]
                    else:
                        result[pairing_type] = cls.load_dict_one_position(json_file_name, pairing_type)
        else:
            # need to create a new file.
            data = {"Watson-Crick": {"A": ["U", "&", "3", "1", "~", "N", "S", ")", "{", "V", "}", "P"],
                                     "C": ["G", "#", "W"], "G": ["C", "B"], "U": ["A"]
                                     },
                    "Wobble": {"A": ["A", "I", "M", "?"], "C": ["A", "U", "P", "I", "?", "Q"], 
                               "G": ["A", "U", "&", "3", "1", "~", "N", "S", ")", "{", "V", "P", "?", "M"],
                               "U": ["G", "#", "W", "U", "V", "P", "I", "Q"]
                               },
                    "Pairing Rules": {
                        "Near-Cognate": {
                            "base-level": [["Wo", "WC", "X"], ["X", "WC", "Wo"]]
                        }
                    }
                    }
            for pairing_type, pairing_data in data.items():
                if pairing_type == "Pairing Rules":
                    result[pairing_type] = pairing_data
                else:
                    result[pairing_type] = cls.convert_one_position_from_dict(data, pairing_type)
        return result


@dataclass
class ClassificationRules:
    classification_type: str
    rules: list[(str, str, str)]


class Window(QMainWindow):

    def __init__(self, pairings, file_name):
        # pylint: disable=invalid-name
        super().__init__()
        self.file_name = file_name
        self.setWindowTitle("Base pairing editor: " + os.path.basename(self.file_name))
        widget = QWidget()
        layout = QGridLayout()

        tree = QTreeWidget()
        tree.setColumnCount(2)
        tree.setHeaderLabels(["Codon", "Anticodons"])
        tree.setColumnWidth(0, 200)

        root = QTreeWidgetItem(tree)
        root.setText(0, "Pairing type")

        self.watson_crick_pairing_widgetItem = QTreeWidgetItem(root)
        self.watson_crick_pairing_widgetItem.setText(0, "Watson-Crick")

        # fill Watson-Crick
        for pairing in pairings["Watson-Crick"]:
            node = QTreeWidgetItem((pairing.codon, ','.join(pairing.anticodons)))
            self.watson_crick_pairing_widgetItem.addChild(node)

        self.wobble_pairing_widgetItem = QTreeWidgetItem(root)
        self.wobble_pairing_widgetItem.setText(0, "Wobble")

        self.pairing_rules = QTreeWidgetItem(tree)
        self.pairing_rules.setText(0, "Pairing Rules")
        self.near_cognate = QTreeWidgetItem(self.pairing_rules)
        self.near_cognate.setText(0, "Near-Cognate")
        self.near_cognate_base_rules = QTreeWidgetItem(("base-level",
                                                        str(pairings["Pairing Rules"]["Near-Cognate"]["base-level"])))
        self.near_cognate.addChild(self.near_cognate_base_rules)

        # fill Wobble
        for pairing in pairings["Wobble"]:
            node = QTreeWidgetItem((pairing.codon, ','.join(pairing.anticodons)))
            self.wobble_pairing_widgetItem.addChild(node)

        tree.itemDoubleClicked.connect(self.edit_rule)

        layout.addWidget(tree, 0, 0)

        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save)
        layout.addWidget(self.save_button, 1, 0)

        widget.setLayout(layout)
        self.setCentralWidget(widget)
        tree.expandAll()

        self.show()

    def identify_click(self, clicked_obj) -> (str, str):
        for child_number in range(self.watson_crick_pairing_widgetItem.childCount()):
            if self.watson_crick_pairing_widgetItem.child(child_number) == clicked_obj:
                return "Watson-Crick"
            if self.wobble_pairing_widgetItem.child(child_number) == clicked_obj:
                return "Wobble"
            if self.near_cognate.child(child_number) == clicked_obj:
                return "Near Cognate"
        return None

    def edit_rule(self, clicked_obj: QTreeWidgetItem):
        pairing_type = self.identify_click(clicked_obj)
        if pairing_type in ["Watson-Crick", "Wobble"]:
            self.edit_pairing(clicked_obj, pairing_type)
        elif pairing_type == "Near Cognate":
            self.edit_near_cognate_definition(clicked_obj)

    def edit_pairing(self, clicked_obj: QTreeWidgetItem, pairing_type: str):
        # open dialog for changing data.
        selected_pairing = PairingRelationship(clicked_obj.data(0, 0), clicked_obj.data(1, 0))
        pairing_dialog = EditPairingDialog(self)
        pairing_dialog.set_pair(selected_pairing)

        pairing_dialog.exec_()
        if pairing_dialog.pairing_relationship.codon == selected_pairing.codon and\
           pairing_dialog.pairing_relationship.anticodons == selected_pairing.anticodons:
            # nothing changed.
            return
        # we need to update the tree.abs
        pairing_node = None
        if pairing_type == "Watson-Crick":
            pairing_node = self.watson_crick_pairing_widgetItem
        else:
            pairing_node = self.wobble_pairing_widgetItem
        for child_no in range(pairing_node.childCount()):
            item = pairing_node.child(child_no)
            if pairing_dialog.pairing_relationship.codon == item.data(0, 0):
                # update
                item.setData(0, 0, pairing_dialog.pairing_relationship.codon)
                item.setData(1, 0, pairing_dialog.pairing_relationship.anticodons)

    def edit_near_cognate_definition(self, clicked_obj):
        # open dialog for changing data.
        near_cognate_def_dialog = EditNearCognateDefinition(self)
        near_cognate_def_dialog.set_rules(clicked_obj.data(1, 0))
        near_cognate_def_dialog.exec_()
        item = self.near_cognate.child(0)
        item.setData(1, 0, near_cognate_def_dialog.rules)

    def save(self):
        # recreate the dictionary
        data = {'Watson-Crick': {}, 'Wobble': {}, "Pairing Rules": {"Near-Cognate": {"base-level": []}}}
        # Watson-Crick:
        for child_no in range(self.watson_crick_pairing_widgetItem.childCount()):
            item = self.watson_crick_pairing_widgetItem.child(child_no)
            data['Watson-Crick'][item.data(0, 0)] = [
                k for k in item.data(1, 0).replace(' ', '').split(',')
            ]

        # Wobble
        for child_no in range(self.wobble_pairing_widgetItem.childCount()):
            item = self.wobble_pairing_widgetItem.child(child_no)
            data['Wobble'][item.data(0, 0)] = [
                k for k in item.data(1, 0).replace(' ', '').split(',')
            ]

        # Near cognate definition
        near_cognate_definition_str = self.near_cognate.child(0).data(1, 0)
        near_cognate_definition = (near_cognate_definition_str.replace('[', '').replace(']', '')
                                   .replace("'", '').split(', '))
        data['Pairing Rules']['Near-Cognate']['base-level'] = [
            [near_cognate_definition[i], near_cognate_definition[i+1], near_cognate_definition[i+2]]
            for i, _ in list(enumerate(near_cognate_definition))[::3]
        ]
        # confirm save the data.
        qm = QMessageBox
        ret = qm.question(self, '', "Are you sure to save the file?", qm.Yes | qm.No)
        if ret == qm.Yes:
            # do save it
            with open(self.file_name, "w", encoding="utf-8") as outfile:
                outfile.write(json.dumps(data, indent=4))


class EditPairingDialog(QDialog):
    def __init__(self, parent=None):
        super(EditPairingDialog, self).__init__(parent)
        self.setWindowTitle("Edit base pairing")
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.clicked_ok)
        self.buttonBox.rejected.connect(self.clicked_cancel)

        self.codon_textbox = QLineEdit(self)
        self.codon_textbox.setInputMask("A")

        self.anticodons_textbox = QLineEdit(self)
        self.anticodons_textbox.setFixedWidth(170)
        # self.validator = QtGui.QRegExpValidator(QtCore.QRegExp("+(\S)(,|$)"), self.anticodons_textbox)
        # self.anticodons_textbox.setValidator(self.validator)

        self.layout = QGridLayout(self)
        self.layout.addWidget(QLabel("Codon"), 0, 0)
        self.layout.addWidget(self.codon_textbox, 0, 1)
        self.layout.addWidget(QLabel("Anticodons"), 1, 0)
        self.layout.addWidget(self.anticodons_textbox, 1, 1)
        self.layout.addWidget(self.buttonBox, 2, 0)

    def clicked_ok(self):
        if all(len(k) == 1 for k in self.anticodons_textbox.text().replace(' ', '').split(',')):
            # update the pairing
            self.pairing_relationship = PairingRelationship(self.codon_textbox.text(), self.anticodons_textbox.text())
            self.close()
        else:
            QMessageBox.about(self, "Invalid input", "The input is invalid: Either there is a codon with a relationship"
                              "or the list of anitcodons is not comma separated. Please review your input.")

    def clicked_cancel(self):
        self.close()

    def set_pair(self, pair):
        self.pairing_relationship = pair
        self.codon_textbox.setText(self.pairing_relationship.codon)
        self.anticodons_textbox.setText(self.pairing_relationship.anticodons)


class EditNearCognateDefinition(QDialog):
    def __init__(self, parent=None):
        super(EditNearCognateDefinition, self).__init__(parent)
        self.setWindowTitle("Edit near-cognate definition")
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.clicked_ok)
        self.buttonBox.rejected.connect(self.clicked_cancel)

        self.definition_rules = QLineEdit(self)
        self.definition_rules.setFixedWidth(350)

        self.layout = QGridLayout(self)
        self.layout.addWidget(QLabel("Near-congnate rules"), 0, 0)
        self.layout.addWidget(self.definition_rules, 0, 1)
        self.layout.addWidget(self.buttonBox, 2, 0)

    def clicked_ok(self):
        # check if it is valid.
        raw_list = (self.definition_rules.text().replace('[', '').
                    replace(']', '').replace("'", '').split(', '))
        if (not [x for x in raw_list if x not in ['Wo', 'WC', 'X']]) and len(raw_list) % 3 == 0:
            # update the pairing
            self.rules = self.definition_rules.text()
            self.close()
        else:
            QMessageBox.about(self, "Invalid input", "The input is invalid: the input should be in the format of "
                                                     "[[pairing_type,pairing_type,pairing_type]], where pairing_type "
                                                     "is either 'WC', 'Wo' or 'X'. More than one "
                                                     "[pairing_type,pairing_type,pairing_type] is allowed.")

    def clicked_cancel(self):
        self.close()

    def set_rules(self, rules):
        self.rules = rules
        self.definition_rules.setText(self.rules)


def main(file_name=None):
    if file_name is None and len(sys.argv) <= 1:
        raise ValueError("Base pairing file not informed.")
    if file_name is None:
        file_name = sys.argv[1]
    app = QApplication(sys.argv)
    pairings = None
    pairings = PairingRelationship.load_all_dict(file_name)
    window = Window(pairings, file_name)
    window.resize(int(window.width()*1.5), int(window.height()*2.5))
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 02 11:32:41 2019

Re-implementation of the R-code that calculates concentrations
This script is intented to generate new concentrations (WC, wobble, near) with different
interpretation of wobble and near cognates. it is intended to be used in GA algorithm.

@author: heday
"""

from importlib import resources
import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import concentrations

def identify_3_base_pairing(codon: str, trna: str, basepairing_rules:dict) -> list[str]:
    """
    Given a codon and a tRNA anticodon, Identify the base pairing type for each base.
    """
    result = []
    for i, c in enumerate(codon):
        anticodon = trna[2 - i]
        if anticodon in basepairing_rules["Watson-Crick"][c]:
            result.append("WC")
        elif anticodon in basepairing_rules["Wobble"][c]:
            result.append("Wo")
        else:
            result.append("X")
    return result

def is_near_cognate(pairing_id:list[str], near_cognate_rules:list[list[str]]) -> bool:
    """
    Given a pairing id (WC, Wo or X), and a list of near-cognate rules,
    identify if this pairing is a near-cognate.
    """
    for partial_rule in near_cognate_rules:
        matches = 0
        for i, pairing_type in enumerate(pairing_id):
            if partial_rule[i].upper() == 'X':
                # doesn't matter, it is always a match.
                matches += 1
            elif pairing_type.upper() == partial_rule[i].upper():
                # match.
                matches += 1
            elif pairing_type.upper() == 'WC' and partial_rule[i].upper() == 'WO':
                # this is also a match.
                matches += 1
        if matches == 3:
            return True
    return False


def make_matrix(t_rnas: pd.DataFrame, codons: pd.DataFrame, verbose=False,
                settings_file_name="default_basepairing.json"):
    """
    Given a DataFrame with tRNA concentrations, and another DataFrame with codons information,
    generates a decoding matrix

    t_rnas: DataFrame with the tRNAs: anticodon, gene.copy.number
    codons: DataFrame with the codons to be used
    verbose: print information about the building stages as it goes
    settings_file_name: json file with the basepairing rules.
    """
    # open file with basepairing rules
    if settings_file_name == "default_basepairing.json":
        with resources.path(concentrations, settings_file_name) as re:
            with open(re, "r", encoding="utf-8") as file:
                basepairing_rules = json.load(file)
    else:
        with open(settings_file_name, "r", encoding="utf-8") as file:
            basepairing_rules = json.load(file)
    # sanity checks
    if "Watson-Crick" not in basepairing_rules.keys():
        raise ValueError("settings file does not contain Watson-Crick pariring rules")
    if "Wobble" not in basepairing_rules.keys():
        raise ValueError("settings file does not contain Wobble pariring rules")
    if "A" not in basepairing_rules["Watson-Crick"]:
        raise ValueError("settings file does not rules for Watson-Crick A basepairing")
    if "C" not in basepairing_rules["Watson-Crick"]:
        raise ValueError("settings file does not rules for Watson-Crick C basepairing")
    if "G" not in basepairing_rules["Watson-Crick"]:
        raise ValueError("settings file does not rules for Watson-Crick G basepairing")
    if "U" not in basepairing_rules["Watson-Crick"]:
        raise ValueError("settings file does not rules for Watson-Crick U basepairing")
    if "A" not in basepairing_rules["Wobble"]:
        raise ValueError("settings file does not rules for Wobble A basepairing")
    if "C" not in basepairing_rules["Wobble"]:
        raise ValueError("settings file does not rules for Wobble C basepairing")
    if "G" not in basepairing_rules["Wobble"]:
        raise ValueError("settings file does not rules for Wobble G basepairing")
    if "U" not in basepairing_rules["Wobble"]:
        raise ValueError("settings file does not rules for Wobble U basepairing")
    # settings seems fine. Proceed.

    # check if tRNAs have 'anticodon' column
    if 'anticodon' not in t_rnas.columns:
        print('tRNA list must contain a column named "anticodon".')
        return
    # check if codons have 'codon' column
    if 'codon' not in codons.columns:
        print('Codon list must contain a column named "codon".')
        return



    cognate_wc_matrix = np.zeros((len(t_rnas.anticodon), len(codons.codon)))
    cognate_wobble_matrix = np.zeros((len(t_rnas.anticodon), len(codons.codon)))
    nearcognate_matrix = np.zeros((len(t_rnas.anticodon), len(codons.codon)))

    if verbose:
        print("Populating WC matrix...")
    # populate cognate WC matrix if WC criteria matched
    for anticodon_index, anticodon in enumerate(t_rnas.anticodon):
        for codon_index, codon in enumerate(codons.codon):
            if identify_3_base_pairing(codon, anticodon, basepairing_rules) == ["WC", "WC", "WC"]:
                cognate_wc_matrix[anticodon_index, codon_index] = 1

    if verbose:
        print("done.")
        print("Populating wobble matrix...")


    # populate cognate wobble matrix
    # if wobble criteria matched, amino acid is correct, and WC matrix entry is 0
    # if incorrect amino acid, assign to near-cognates
    for anticodon_index, anticodon in enumerate(t_rnas.anticodon):
        for codon_index, codon in enumerate(codons.codon):
            if cognate_wc_matrix[anticodon_index,codon_index] == 0 and\
               identify_3_base_pairing(codon, anticodon, basepairing_rules) == ["WC", "WC", "Wo"]:
                if t_rnas["three.letter"][anticodon_index] == codons["three.letter"][codon_index]:
                    cognate_wobble_matrix[anticodon_index,codon_index] = 1
                else:
                    nearcognate_matrix[anticodon_index,codon_index] = 1

    if verbose:
        print('done.')
        print('Populating nearcognate matrix...')

    #populate near-cognate matrix if:
    #wobble and WC matrix entries are 0,
    #wobble criteria are matched

    for anticodon_index, _ in enumerate(t_rnas.anticodon):
        for codon_index, _ in enumerate(codons.codon):
            if (cognate_wc_matrix[anticodon_index,codon_index] == 0 and\
                cognate_wobble_matrix[anticodon_index,codon_index] == 0) and\
                is_near_cognate(identify_3_base_pairing(codons.codon[codon_index],
                                                        t_rnas.anticodon[anticodon_index],
                                                        basepairing_rules),
                                basepairing_rules["Pairing Rules"]["Near-Cognate"]["base-level"]):
                nearcognate_matrix[anticodon_index,codon_index] = 1

    if verbose:
        print('done.')

    #Sanity checks

    #  Check whether any tRNA:codon combination is assigned 1
    #  in more than one table (this should not occur)

    testsum = cognate_wc_matrix + cognate_wobble_matrix + nearcognate_matrix
    if np.any(testsum>1):
        print('Warning: multiple relationships for identical tRNA:codon pairs detected.')
        return {}
    if verbose:
        print('No threesome errors detected.')

    return {"cognate.wc.matrix":cognate_wc_matrix, "cognate.wobble.matrix":cognate_wobble_matrix,
            "nearcognate.matrix":nearcognate_matrix}

def plot_matrix(matrices_dict: dict, t_rnas: pd.DataFrame, codons: pd.DataFrame, save_fig = None):
    """
    Plots the pairing matrices.
    """
    colours=['g', 'y', 'r']
    labels = list(matrices_dict.keys())
    i = 0
    plt.figure(figsize=(25,15))
    plt.grid(True)
    for k in matrices_dict.keys():
        matches_dict = np.argwhere(matrices_dict[k] == 1)#[:,1]
        # display(c)
        plt.plot(matches_dict[:,1], matches_dict[:,0], colours[i] + 's', label=labels[i])
        i +=1
    plt.xticks(range(len(codons.codon)), codons.codon, rotation = 45)
    plt.yticks(range(len(t_rnas.anticodon)), t_rnas.anticodon)
    plt.legend()
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()

def print_codon_anticodon_pairings(matrices_dict: dict, t_rnas: pd.DataFrame, codons: pd.DataFrame):
    """
    prints the pairing of all codons with all anticodons.
    """
    for k in matrices_dict.keys():
        print(" Pairings for: " + k)
        matches_dict = np.argwhere(matrices_dict[k] == 1)
        for anticodon, c in matches_dict:
            print (codons.codon[c] + " - " + t_rnas.anticodon[anticodon])

def make_concentrations(matrices_dict: dict, t_rnas: pd.DataFrame, codons: pd.DataFrame,
                        concentration_col_name = 'gene.copy.number', total_trna=190, verbose=False):
    """
    Given a tRNA matrix, and the decoding matrix, generates a concentrations DataFrame.

    TRNAs: DataFrame with the tRNAs: anticodon, gene.copy.number
    Matrices: pairing matrices generated by make_matrix
    Codons: DataFrame with the codons to be used
    concentration_col_name: name of the concentrations column name. Default = 'gene.copy.number'
    total_Trna: default value is 190 (micromoles).
    """
    wc_cognate = matrices_dict["cognate.wc.matrix"]
    wobblecognate = matrices_dict["cognate.wobble.matrix"]
    nearcognate = matrices_dict["nearcognate.matrix"]

    # construct empty results dataframe
    trna_concentrations = pd.DataFrame(codons[[codons.columns[0], codons.columns[1]]])
    trna_concentrations["WCcognate.conc"] = 0.0
    trna_concentrations["wobblecognate.conc"] = 0.0
    trna_concentrations["nearcognate.conc"] = 0.0

    # calculate a conversion factor to convert the abundance factor to a molar concentration
    if verbose:
        print('using: '+ concentration_col_name)
    conversion_factor = np.float64(total_trna /
                                   np.float64(t_rnas[concentration_col_name].sum()) * 1e-6)

    # go through the WCcognates matrix and for each entry of 1 add the abundance of the tRNA
    # from the abundance table to the concentration table
    for codon_index, _ in enumerate(codons.codon):
        for anticodon_index, _ in enumerate(t_rnas.anticodon):
            if wc_cognate[anticodon_index, codon_index] == 1:
                trna_concentrations.loc[codon_index, "WCcognate.conc"] =\
                    trna_concentrations["WCcognate.conc"][codon_index] +\
                        (t_rnas[concentration_col_name][anticodon_index]*conversion_factor)

    #ditto for wobblecognate
    for codon_index, _ in enumerate(codons.codon):
        for anticodon_index, _ in enumerate(t_rnas.anticodon):
            if wobblecognate[anticodon_index, codon_index] == 1:
                trna_concentrations.loc[codon_index, "wobblecognate.conc"] =\
                    trna_concentrations["wobblecognate.conc"][codon_index] +\
                        (t_rnas[concentration_col_name][anticodon_index]*conversion_factor)

    #ditto for nearcognates
    for codon_index, _ in enumerate(codons.codon):
        for anticodon_index, _ in enumerate(t_rnas.anticodon):
            if nearcognate[anticodon_index, codon_index] == 1:
                trna_concentrations.loc[codon_index, "nearcognate.conc"] =\
                    trna_concentrations["nearcognate.conc"][codon_index] +\
                        (t_rnas[concentration_col_name][anticodon_index]*conversion_factor)
    return trna_concentrations

##example of how to use:

# tRNAs = pd.read_csv('/home/heday/Projects/R_concentrations/data/tRNAs.csv')
# codons = pd.read_csv('/home/heday/Projects/R_concentrations/data/codons.csv')
# matrix = make_matrix(tRNAs, codons, verbose=True)
# make_concentrations(matrix, tRNAs, codons, verbose=True)



'''
Utility functions
'''

import numpy as np

def get_number_elongating_ribosomes_per_time(sim):
    """
    For each entry in the simulation log, returns the number of elongating ribosomes.
    """
    return [len(entry) for entry in sim.ribosome_positions_history], np.cumsum(sim.dt_history)


def get_codon_average_occupancy(sim):
    '''
    Returns a tuple where the first element is a vector with the
    enlogation duration of the ribosomes that terminated in the simulation, and
    the second element is a vector with the iteration where such ribosomes
    started enlogating.
    '''
    # sim.updateRibosomeHistory()
    return sim.getEnlogationDuration()



""" instaler """
# -*- coding: utf-8 -*-
# :Project:   elongation_simulators -- Packaging
# :Author:    Fabio Hedayioglu <fheday@gmail.com>
# :License:   MIT License
# :Copyright: © 2020 Fabio Hedayioglu
#

import sys
import sysconfig
from setuptools import setup
from setuptools.command.build_ext import build_ext
from pybind11.setup_helpers import Pybind11Extension,\
                            ParallelCompile  # noqa:E402


WIN = sys.platform.startswith("win32") and "mingw" not in sysconfig.get_platform()

# Avoid a gcc warning below:
# cc1plus: warning: command line option ‘-Wstrict-prototypes’ is valid
# for C/ObjC but not for C++


class BuildExt(build_ext):
    """ Class to remove warning on gcc."""
    def build_extensions(self):
        if '-Wstrict-prototypes' in self.compiler.compiler_so:
            self.compiler.compiler_so.remove('-Wstrict-prototypes')
        super().build_extensions()


EXTRA_COMPILE_ARGS = None
if WIN:
    EXTRA_COMPILE_ARGS = ["/std:c++17", "/O2", "/Ot", "/GL", "/DCOMIPLE_PYTHON_MODULE", "/I./src/eigen-3.3.7/eigen3/"]
else:
    EXTRA_COMPILE_ARGS = ["-O3", "-ffast-math", "-ftree-vectorize", "-Wall",
                          "-g2", "-flto=auto", "-DCOMIPLE_PYTHON_MODULE"]

ext_modules = [
    Pybind11Extension(
        "sequence_simulator",
        ["src/concentrationsreader.cpp", "src/mrna_reader.cpp", "src/elongation_codon.cpp",
         "src/initiationterminationcodon.cpp", "src/mrnaelement.cpp", "src/sequence_simulator.cpp",
         "src/codon_simulator.cpp", "src/elongation_simulation_manager.cpp",
         "src/elongation_simulation_processor.cpp", "./src/jsoncpp/jsoncpp.cpp"],
        include_dirs=["./src/jsoncpp/", "./src/eigen-3.3.7/"],
        extra_compile_args=EXTRA_COMPILE_ARGS
    ),
    Pybind11Extension(
        "codon_simulator",
        ["src/concentrationsreader.cpp", "src/mrna_reader.cpp", "src/codon_simulator.cpp",
         "./src/jsoncpp/jsoncpp.cpp"],
        include_dirs=["./src/jsoncpp/", "./src/eigen-3.3.7/"],
        extra_compile_args=EXTRA_COMPILE_ARGS
    )
]


if sys.version_info < (3,):
    raise NotImplementedError("Only Python 3+ is supported.")


with open('version.txt', encoding='utf-8') as f:
    VERSION = f.read()

with open('README.rst', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

with open('CHANGES.rst', encoding='utf-8') as f:
    CHANGES = f.read()

# Optional multithreaded build
ParallelCompile("NPY_NUM_BUILD_JOBS").install()

CMDCLASS = {}
if not WIN:
    CMDCLASS = {"build_ext": BuildExt}


setup(
    name='elongation_simulator',
    version=VERSION,
    description='High-performance Codon simulator"\
     " and elongation simulator for eukaryotic organism',
    long_description=LONG_DESCRIPTION + '\n\n' + CHANGES,
    # long_description_content_type='text/x-rst',
    license='MIT License',
    keywords='elongation translation',
    author='Fabio Hedayioglu',
    author_email='fheday@gmail.com',
    maintainer='Fabio Hedayioglu',
    maintainer_email='fheday@gmail.com',
    url='https://github.com/fheday/elongation_simulator/',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python',
    ],
    install_requires=['setuptools', 'pybind11', 'pytest', 'numpy', 'pyqt5'],
    packages=["concentrations", "elongation"],
    package_dir={"concentrations": "concentrations", "elongation": "elongation"},
    scripts=['concentrations/basepairingeditor.py', 'elongation/simulationbuilder.py'],
    cmdclass=CMDCLASS,
    ext_modules=ext_modules,
    zip_safe=False,
    include_package_data=True
)
